@@ -920,6 +920,30 @@ generic_delete(int valid, MDB_txn *txn, DbObject *db,
     Py_RETURN_TRUE;
 }
 
+static PyObject *trans_abort(TransObject *);
+static PyObject *trans_commit(TransObject *);
+
+static PyObject *
+generic_finish(TransObject *trans, PyObject *ret)
+{
+    if(PyErr_Occurred()) {
+        Py_CLEAR(ret);
+    }
+    PyObject *rett;
+    if(ret) {
+        rett = trans_commit(trans);
+    } else {
+        rett = trans_abort(trans);
+    }
+    Py_DECREF(trans);
+    if(! rett) {
+        Py_CLEAR(ret);
+        return NULL;
+    }
+    Py_DECREF(rett);
+    return ret;
+}
+
 static PyObject *
 make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buffers)
 {
@@ -1379,7 +1403,6 @@ env_info(EnvObject *self)
     return dict_from_fields(&info, fields);
 }
 
-
 static PyObject *
 env_open_db(EnvObject *self, PyObject *args, PyObject *kwds)
 {
@@ -1421,7 +1444,6 @@ env_open_db(EnvObject *self, PyObject *args, PyObject *kwds)
     }
 }
 
-
 static PyObject *
 env_path(EnvObject *self)
 {
@@ -1437,7 +1459,6 @@ env_path(EnvObject *self)
     return PyUnicode_FromString(path);
 }
 
-
 static const struct dict_field mdb_stat_fields[] = {
     {TYPE_UINT, "psize",          offsetof(MDB_stat, ms_psize)},
     {TYPE_UINT, "depth",          offsetof(MDB_stat, ms_depth)},
@@ -1448,7 +1469,6 @@ static const struct dict_field mdb_stat_fields[] = {
     {TYPE_EOF, NULL, 0}
 };
 
-
 static PyObject *
 env_stat(EnvObject *self)
 {
@@ -1466,6 +1486,51 @@ env_stat(EnvObject *self)
     return dict_from_fields(&st, mdb_stat_fields);
 }
 
+static int env_readers_callback(const char *msg, void *str_)
+{
+    PyObject **str = str_;
+    int old_size = PyString_GET_SIZE(*str);
+    int chunk_size = strlen(msg);
+
+    if(_PyString_Resize(str, old_size + chunk_size)) {
+        return -1;
+    }
+    memcpy(PyString_AS_STRING(*str) + old_size, msg, chunk_size);
+    return 0;
+}
+
+static PyObject *
+env_readers(EnvObject *self)
+{
+    if(! self->valid) {
+        return err_invalid();
+    }
+
+    PyObject *str = PyString_FromStringAndSize(NULL, 20);
+    if(! str) {
+        return NULL;
+    }
+
+    if(mdb_reader_list(self->env, env_readers_callback, &str)) {
+        Py_CLEAR(str);
+    }
+    return str;
+}
+
+static PyObject *
+env_reader_check(EnvObject *self)
+{
+    if(! self->valid) {
+        return err_invalid();
+    }
+
+    int dead;
+    int rc = mdb_reader_check(self->env, &dead);
+    if(rc) {
+        return err_set("mdb_reader_check", rc);
+    }
+    return PyInt_FromLong(dead);
+}
 
 static PyObject *
 env_sync(EnvObject *self, PyObject *args)
@@ -1579,32 +1644,6 @@ env_gets(EnvObject *self, PyObject *args, PyObject *kwds)
     return dict;
 }
 
-
-static PyObject *trans_abort(TransObject *);
-static PyObject *trans_commit(TransObject *);
-
-static PyObject *
-generic_finish(TransObject *trans, PyObject *ret)
-{
-    if(PyErr_Occurred()) {
-        Py_CLEAR(ret);
-    }
-    PyObject *rett;
-    if(ret) {
-        rett = trans_commit(trans);
-    } else {
-        rett = trans_abort(trans);
-    }
-    Py_DECREF(trans);
-    if(! rett) {
-        Py_CLEAR(ret);
-        return NULL;
-    }
-    Py_DECREF(rett);
-    return ret;
-}
-
-
 static PyObject *
 env_put(EnvObject *self, PyObject *args, PyObject *kwds)
 {
@@ -1617,7 +1656,6 @@ env_put(EnvObject *self, PyObject *args, PyObject *kwds)
     return generic_finish(trans, ret);
 }
 
-
 static PyObject *
 env_puts(EnvObject *self, PyObject *args, PyObject *kwds)
 {
@@ -1838,6 +1876,8 @@ static struct PyMethodDef env_methods[] = {
     {"open_db", (PyCFunction)env_open_db, METH_VARARGS|METH_KEYWORDS},
     {"path", (PyCFunction)env_path, METH_NOARGS},
     {"stat", (PyCFunction)env_stat, METH_NOARGS},
+    {"readers", (PyCFunction)env_readers, METH_NOARGS},
+    {"reader_check", (PyCFunction)env_reader_check, METH_NOARGS},
     {"sync", (PyCFunction)env_sync, METH_VARARGS},
     {"get", (PyCFunction)env_get, METH_VARARGS|METH_KEYWORDS},
     {"gets", (PyCFunction)env_gets, METH_VARARGS|METH_KEYWORDS},
@@ -2666,15 +2706,6 @@ PyTypeObject PyTransaction_Type = {
 };
 
 
-static int add_type(PyObject *mod, PyTypeObject *type)
-{
-    if(PyType_Ready(type)) {
-        return -1;
-    }
-    return PyObject_SetAttrString(mod, type->tp_name, (PyObject *)type);
-}
-
-
 static PyObject *
 enable_drop_gil(void)
 {
@@ -2704,6 +2735,74 @@ static struct PyModuleDef moduledef = {
 #endif
 
 
+/**
+ * Initialize and publish the LMDB built-in types.
+ */
+static int init_types(PyObject *mod)
+{
+    static PyTypeObject *types[] = {
+        &PyEnvironment_Type,
+        &PyCursor_Type,
+        &PyTransaction_Type,
+        &PyIterator_Type,
+        &PyDatabase_Type,
+        NULL
+    };
+
+    int i;
+    for(i = 0; types[i]; i++) {
+        PyTypeObject *type = types[i];
+        if(PyType_Ready(type)) {
+            return -1;
+        }
+        if(PyObject_SetAttrString(mod, type->tp_name, (PyObject *)type)) {
+            return -1;
+        }
+    }
+    return 0;
+}
+
+
+/**
+ * Produce `string_tbl' array of PyObjects from `strings'.
+ */
+static int init_strings(PyObject *mod)
+{
+    string_tbl = malloc(sizeof(PyObject *) * STRING_ID_COUNT);
+    if(! string_tbl) {
+        return -1;
+    }
+
+    const char *cur = strings;
+    int i;
+    for(i = 0; i < STRING_ID_COUNT; i++) {
+        if(! ((string_tbl[i] = PyUnicode_InternFromString(cur)))) {
+            return -1;
+        }
+        cur += PyString_GET_SIZE(string_tbl[i]) + 1;
+    }
+    return 0;
+}
+
+
+/**
+ * Initialize a bunch of constants used to ease number compares.
+ */
+static int init_constants(PyObject *mod)
+{
+    if(! ((py_zero = PyLong_FromUnsignedLongLong(0)))) {
+        return -1;
+    }
+    if(! ((py_int_max = PyLong_FromUnsignedLongLong(INT_MAX)))) {
+        return -1;
+    }
+    if(! ((py_size_max = PyLong_FromUnsignedLongLong(SIZE_MAX)))) {
+        return -1;
+    }
+    return 0;
+}
+
+
 PyMODINIT_FUNC
 MODINIT_NAME(void)
 {
@@ -2716,41 +2815,13 @@ MODINIT_NAME(void)
         MOD_RETURN(NULL);
     }
 
-    static PyTypeObject *types[] = {
-        &PyEnvironment_Type,
-        &PyCursor_Type,
-        &PyTransaction_Type,
-        &PyIterator_Type,
-        &PyDatabase_Type,
-        NULL
-    };
-    int i;
-    for(i = 0; types[i]; i++) {
-        if(add_type(mod, types[i])) {
-            MOD_RETURN(NULL);
-        }
-    }
-
-    string_tbl = malloc(sizeof(PyObject *) * STRING_ID_COUNT);
-    if(! string_tbl) {
+    if(init_types(mod)) {
         MOD_RETURN(NULL);
     }
-
-    const char *cur = strings;
-    for(i = 0; i < STRING_ID_COUNT; i++) {
-        if(! ((string_tbl[i] = PyUnicode_InternFromString(cur)))) {
-            MOD_RETURN(NULL);
-        }
-        cur += strlen(cur) + 1;
-    }
-
-    if(! ((py_zero = PyLong_FromUnsignedLongLong(0)))) {
+    if(init_strings(mod)) {
         MOD_RETURN(NULL);
     }
-    if(! ((py_int_max = PyLong_FromUnsignedLongLong(INT_MAX)))) {
-        MOD_RETURN(NULL);
-    }
-    if(! ((py_size_max = PyLong_FromUnsignedLongLong(SIZE_MAX)))) {
+    if(init_constants(mod)) {
         MOD_RETURN(NULL);
     }
 
